/* A-826 service module.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

  v 0.0.0 29 Apr 2003 by Reed Lai
    create, blah blah... */

/* Mandatory */
#include <linux/kernel.h>		/* ta, kernel work */
#include <linux/module.h>		/* is a module */

/* Deal with CONFIG_MODVERSIONS that is defined in
   /usr/include/linux/config.h (config.h is included by module.h) */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* Additional */
#include <linux/fs.h>
	/* file system here */
#include <linux/wrapper.h>
	/* for compatibility with future version of Linux */

/* use I/O ports */
#include <asm/io.h>
#include <linux/ioport.h>

/* need kmalloc */
// #include <linux/slab.h>

/* irq service */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

/* wow, dma */
#include <asm/dma.h>

/* Local */
#include "ixisa.h"
#include "_flag_inline.h"
#include "_a826.h"

#define BUF_LEN 80

char msg[BUF_LEN];
char *msg_ptr;

ixisa_devinfo_t *dev;

/* symbols exportation */
// EXPORT_NO_SYMBOLS; /* do not export symbols */
EXPORT_SYMBOL_NOVERS(ixa826_ioctl);
EXPORT_SYMBOL_NOVERS(ixa826_release);
EXPORT_SYMBOL_NOVERS(ixa826_open);

static __inline__ void clear_board_irq(unsigned long base)
{
#define WHATEVER 0
	outb(WHATEVER, base + CIR);
}

static __inline__ void config_8254_count(unsigned long base, char id,
										 unsigned char cw,
										 unsigned int val)
{
	outb(cw, base + _8254CR);
	outb(val, base + id);
	outb(val >> 8, base + id);
}

static __inline__ void config_board_pacer_count(unsigned long base,
												unsigned int val)
{
	unsigned int c1, c2;

	c1 = (val >> 16) & 0xffff;
	c2 = val & 0xffff;
	if (c1 == 0) {
		c1 = 2;					/* the counter value 1 in mode 2 is illegal */
		c2 = c2 >> 1;
	}

	/* Mode 2 - rate generator */
	config_8254_count(base, 1,
					  _8254CW_COUNTER_1 | _8254CW_RW_LMB | _8254CW_MODE_2,
					  c1);
	config_8254_count(base, 2,
					  _8254CW_COUNTER_2 | _8254CW_RW_LMB | _8254CW_MODE_2,
					  c2);
}

static __inline__ void stop_board_pacer_count(unsigned long base)
{
	/* counter 1: mode 0 */
	config_8254_count(base, 1,
					  _8254CW_COUNTER_1 | _8254CW_RW_LMB | _8254CW_MODE_0,
					  0xffff);

	/* counter 2: mode 0 */
	config_8254_count(base, 2,
					  _8254CW_COUNTER_2 | _8254CW_RW_LMB | _8254CW_MODE_0,
					  0xffff);
}

void reset_board_count(ixisa_devinfo_t * dp)
{
	/* counter 0: mode 0 */
	config_8254_count(dp->base, 0,
					  _8254CW_COUNTER_0 | _8254CW_RW_LMB | _8254CW_MODE_0,
					  0xffff);

	/* counter 1: mode 0 */
	config_8254_count(dp->base, 1,
					  _8254CW_COUNTER_1 | _8254CW_RW_LMB | _8254CW_MODE_0,
					  0xffff);

	/* counter 2: mode 0 */
	config_8254_count(dp->base, 2,
					  _8254CW_COUNTER_2 | _8254CW_RW_LMB | _8254CW_MODE_0,
					  0xffff);
}

static __inline__ void disable_board_dma(ixisa_devinfo_t * dp)
{
#if ADMCR_MASK == 0xff
	outb(((unsigned char) dp->admcr), dp->base + ADMCR);
#else
	outb(((unsigned char) dp->admcr) & ADMCR_MASK, dp->base + ADMCR);
#endif
}

static __inline__ void enable_board_dma(unsigned long base)
{
	outb(DMA_TRANSFER_MODE, base + ADMCR);
}

int config_dma(ixisa_devinfo_t * dp)
{
	unsigned long flags;

	if (dp->dma.nr && dp->buf16->buf && dp->buf16->sz) {
		dp->dma.sz = dp->buf16->sz << 1;	/* size in 8-bit */
		dp->dma.buf =
			(void *) __get_dma_pages(GFP_KERNEL, get_order(dp->dma.sz));
		if (!dp->dma.buf)
			return FAILURE;
		flags = claim_dma_lock();
		disable_dma(dp->dma.nr);
		clear_dma_ff(dp->dma.nr);
		set_dma_mode(dp->dma.nr, dp->dma.mode);
		set_dma_addr(dp->dma.nr, virt_to_bus(dp->dma.buf));
		set_dma_count(dp->dma.nr, dp->dma.sz);
		enable_dma(dp->dma.nr);
		release_dma_lock(flags);
	} else
		return FAILURE;

	return SUCCESS;
}

void unconfig_dma(ixisa_devinfo_t * dp)
{
	unsigned long flags;

	flags = claim_dma_lock();
	disable_dma(dp->dma.nr);
	clear_dma_ff(dp->dma.nr);
	release_dma_lock(flags);
	free_pages((unsigned long) dp->dma.buf, get_order(dp->dma.sz));
	dp->dma.buf = 0;
	dp->dma.sz = 0;
}

void reset_board(ixisa_devinfo_t * dp)
{
	reset_board_count(dp);

	dp->admcr = 1;
	disable_board_dma(dp);

	clear_board_irq(dp->base);
}

#if BOTTOM_HELF
void irq_handler_bh(void *data)
{
	/* bottom helf of irq handler
	 *
	 * Arguments:
	 *   data     data pointer as <linux/tqueue.h> defined
	 *
	 * Returned:  none */

	ixisa_devinfo_t *dp;

	dp = (ixisa_devinfo_t *) data;

	/* reserved */
}
#endif							/* BOTTOM_HELF */

void irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* Arguments:  as <linux/sched.h> defined
	 *
	 * Returned:  none */
	ixisa_devinfo_t *dp;

	dp = (ixisa_devinfo_t *) dev_id;

	if (get_flag(&dp->flags, IXISA_FLAG_DMA)) {
		//while (get_dma_residue(dp->dma.nr));
		disable_board_dma(dp);
	}
	clear_board_irq(dp->base);

#if BOTTOM_HELF
	static struct tq_struct task = { NULL, 0, irq_handler_bh, NULL };
	task.data = dev_id;
#endif							/* BOTTOM_HELF */

	send_sig(dp->sig.sid, dp->sig.task, 1);

	/* Scheduale bottom half to run */
#if BOTTOM_HELF
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
	queue_task(&task, &tq_immediate);
#else
	queue_task_irq(&task, &tq_immediate);
#endif
	mark_bh(IMMEDIATE_BH);
#endif							/* BOTTOM_HELF */
}

int set_signal(ixisa_signal_t * sig, ixisa_devinfo_t * dp)
{
	/* set user signal for interrupt
	 *
	 * Arguments:
	 *   sig      pointer to structure ixisa_signal for user signal
	 *   di       pointer to structure ixisa_devinfo for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	if (sig->sid) {
		/* has signal id, set */
		if (sig->pid) {
			/* user gives his process id */
			dp->sig.sid = sig->sid;
			dp->sig.pid = sig->pid;
			dp->sig.task = find_task_by_pid(sig->pid);
		} else if (sig->task) {
			/* user gives his task structure */
			dp->sig.sid = sig->sid;
			dp->sig.task = sig->task;
		} else {
			/* wrong! */
			return FAILURE;
		}
		dp->sig.is = sig->is;
		dp->sig.edge = sig->edge;
		dp->sig.bedge = sig->bedge;
	} else {
		/* no signal id, clear */
		dp->sig.sid = 0;
		dp->sig.pid = 0;
		dp->sig.task = 0;
		dp->sig.is = 0;
		dp->sig.edge = 0;
		dp->sig.bedge = 0;
	}
	return SUCCESS;
}

int write_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Write to register
	 *
	 * Arguments:
	 *   reg      pointer to a structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	unsigned char lbyte, hbyte;

	switch (reg->id) {
	case IXISA_8254C0:
		/* 8254 Counter 0 */
#if _8254C0_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C0);
#else
		outb(((unsigned char) reg->vlaue) & _8254C0_MASK, base + _8254C0);
#endif
		break;
	case IXISA_8254C1:
		/* 8254 Counter 1 */
#if _8254C1_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C1);
#else
		outb(((unsigned char) reg->vlaue) & _8254C1_MASK, base + _8254C1);
#endif
		break;
	case IXISA_8254C2:
		/* 8254 Counter 2 */
#if _8254C2_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254C2);
#else
		outb(((unsigned char) reg->vlaue) & _8254C2_MASK, base + _8254C2);
#endif
		break;
	case IXISA_8254CR:
		/* 8254 Counter Control Register */
#if _8254CR_MASK == 0xff
		outb(((unsigned char) reg->value), base + _8254CR);
#else
		outb(((unsigned char) reg->vlaue) & _8254CR_MASK, base + _8254CR);
#endif
		break;
	case IXISA_AO0L:
#if AO0L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO0L);
#else
		outb(((unsigned char) reg->value) & AO0L_MASK, base + AO0L);
#endif
		break;
	case IXISA_AO0H:
#if AO0H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO0H);
#else
		outb(((unsigned char) reg->value) & AO0H_MASK, base + AO0H);
#endif
		break;
	case IXISA_AO0:
		/* Analog Output (D/A) Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO0L_MASK == 0xff
		outb(lbyte, base + AO0L);
#else
		outb(lbyte & AO0L_MASK, base + AO0L);
#endif
#if AO0H_MASK == 0xff
		outb(hbyte, base + AO0H);
#else
		outb(hbyte & AO0H_MASK, base + AO0H);
#endif
		break;
	case IXISA_AO1L:
#if AO1L_MASK == 0xff
		outb((unsigned char) reg->value, base + AO1L);
#else
		outb(((unsigned char) reg->value) & AO1L_MASK, base + AO1L);
#endif
		break;
	case IXISA_AO1H:
#if AO1H_MASK == 0xff
		outb((unsigned char) reg->value, base + AO1H);
#else
		outb(((unsigned char) reg->value) & AO1H_MASK, base + AO1H);
#endif
		break;
	case IXISA_AO1:
		/* D/A Output Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if AO1L_MASK == 0xff
		outb(lbyte, base + AO1L);
#else
		outb(lbyte & AO1L_MASK, base + DA10L);
#endif
#if AO1H_MASK == 0xff
		outb(hbyte, base + AO1H);
#else
		outb(hbyte & AO1H_MASK, base + AO1H);
#endif
		break;
	case IXISA_CIR:
		/* Clear Interrupt Request */
#if CIR_MASK == 0xff
		outb(((unsigned char) reg->value), base + CIR);
#else
		outb(((unsigned char) reg->value) & CIR_MASK, base + CIR);
#endif
		break;
	case IXISA_ADGCR:
		/* A/D Gain Control Register */
#if ADGCR_MASK == 0xff
		outb(((unsigned char) reg->value), base + ADGCR);
#else
		outb(((unsigned char) reg->value) & ADGCR_MASK, base + ADGCR);
#endif
		break;
	case IXISA_ADMXCR:
		/* A/D Multiplex Control Register */
#if ADMXCR_MASK == 0xff
		outb(((unsigned char) reg->value), base + ADMXCR);
#else
		outb(((unsigned char) reg->value) & ADMXCR_MASK, base + ADMXCR);
#endif
		break;
	case IXISA_ADMCR:
		/* A/D Mode Control Register */
#if ADMCR_MASK == 0xff
		outb(((unsigned char) reg->value), base + ADMCR);
#else
		outb(((unsigned char) reg->value) & ADMCR_MASK, base + ADMCR);
#endif
		break;
	case IXISA_ADSTCR:
		/* A/D Software Trigger Control Register */
#if ADSTCR_MASK == 0xff
		outb(((unsigned char) reg->value), base + ADSTCR);
#else
		outb(((unsigned char) reg->value) & ADSTCR_MASK, base + ADSTCR);
#endif
		break;
	case IXISA_DOL:
		/* D/O Output Latch Register, Low Byte */
#if DOL_MASK == 0xff
		outb(((unsigned char) reg->value), base + DOL);
#else
		outb(((unsigned char) reg->value) & DOL_MASK, base + DOL);
#endif
		break;
	case IXISA_DOH:
		/* D/O Output Latch Register, High Byte */
#if DOH_MASK == 0xff
		outb(((unsigned char) reg->value), base + DOH);
#else
		outb(((unsigned char) reg->value) & DOH_MASK, base + DOH);
#endif
		break;
	case IXISA_DO:
		/* D/O Output Latch Register */
		lbyte = (unsigned char) reg->value;
		hbyte = (unsigned char) (reg->value >> 8);
#if DOL_MASK == 0xff
		outb(lbyte, base + DOL);
#else
		outb(lbyte & DOL_MASK, base + DOL);
#endif
#if DOH_MASK == 0xff
		outb(hbyte, base + DOH);
#else
		outb(hbyte & DOH_MASK, base + DOH);
#endif
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

int read_reg(ixisa_reg_t * reg, unsigned long int base)
{
	/* Read from register
	 *
	 * Arguments:
	 *   reg      pointer to structure ixisa_reg for register
	 *   base     base address for current device
	 *
	 * Returned:  SUCCESS or FAILURE */

	unsigned int lbyte, hbyte;

	switch (reg->id) {
	case IXISA_8254C0:
		/* 8254 Counter 0 */
#if _8254C0_MASK == 0xff
		reg->value = inb(base + _8254C0);
#else
		reg->value = inb(base + _8254C0) & _8254C0_MASK;
#endif
		break;
	case IXISA_8254C1:
		/* 8254 Counter 1 */
#if _8254C1_MASK == 0xff
		reg->value = inb(base + _8254C1);
#else
		reg->value = inb(base + _8254C1) & _8254C1_MASK;
#endif
		break;
	case IXISA_8254C2:
		/* 8254 Counter 2 */
#if _8254C2_MASK == 0xff
		reg->value = inb(base + _8254C2);
#else
		reg->value = inb(base + _8254C2) & _8254C2_MASK;
#endif
		break;
	case IXISA_AIL:
		/* A/D Input Buffer Register, Low Byte */
#if AIL_MASK == 0xff
		reg->value = inb(base + AIL);
#else
		reg->value = inb(base + AIL) & AIL_MASK;
#endif
		break;
	case IXISA_AIH:
		/* A/D Input Buffer Register, High Byte */
#if AIH_MASK == 0xff
		reg->value = inb(base + AIH);
#else
		reg->value = inb(base + AIH) & AIH_MASK;
#endif
		break;
	case IXISA_AI:
		/* A/D Inputer Buffer Register */
		switch (reg->mode) {
		case IXISA_RM_TRIGGER:
			/* send software trigger */
			outb(0, base + ADSTCR);
		case IXISA_RM_READY:
			/* read when ready */
			while (inb(base + ADCR) & AI_READY_BIT) {
				/* FIXME - will cause system hang here if ADC status failure */
			}

		default:
			hbyte = inb(base + AIH);
		}
		lbyte = inb(base + AIL);
		reg->value = (hbyte << 8) + lbyte;
		break;
	case IXISA_ADCR:
		/* A/D Conversion Ready */
#if ADCR_MASK == 0xff
		reg->value = inb(base + ADCR);
#else
		reg->value = inb(base + ADCR) & ADCR_MASK;
#endif
		break;
	case IXISA_DIL:
		/* D/I Input Buffer Register, Low Byte */
#if DIL_MASK == 0xff
		reg->value = inb(base + DIL);
#else
		reg->value = inb(base + DIL) & DIL_MASK;
#endif
		break;
	case IXISA_DIH:
		/* D/I Input Buffer Register, High Byte */
#if DIH_MASK == 0xff
		reg->value = inb(base + DIH);
#else
		reg->value = inb(base + DIH) & DIH_MASK;
#endif
		break;
	case IXISA_DI:
		/* D/I Inputer Buffer Register */
#if DIL_MASK == 0xff
		lbyte = inb(base + DIL);
#else
		lbyte = inb(base + DIL) & DIL_MASK;
#endif
#if DIH_MASK == 0xff
		hbyte = inb(base + DIH);
#else
		hbyte = inb(base + DIH) & DIH_MASK;
#endif
		reg->value = (hbyte << 8) + lbyte;
		break;
	default:
		return FAILURE;
	}
	return SUCCESS;
}

ixisa_devinfo_t *align_to_minor(int minor)
{
	ixisa_devinfo_t *dp;
	for (dp = dev; dp && dp->no != minor && dp->next_f; dp = dp->next_f);
	if (dp->no == minor)
		return dp;
	return 0;
}

int ixa826_ioctl(struct inode *inode, struct file *file,
				 unsigned int ioctl_num, unsigned long ioctl_param)
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process tries to do and
	 * IO control on PIO device file
	 *
	 * Arguments: read <linux/fs.h> for (*ioctl) of struct file_operations
	 *
	 * Returned:  SUCCESS or FAILED */

	ixisa_devinfo_t *dp;
	//unsigned int irq_bits;

	dp = align_to_minor(MINOR(inode->i_rdev));

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	if (!dp->open) {
		/* if not yet opened, get out */
		return -EINVAL;
	}

	switch (ioctl_num) {
	case IXISA_GET_INFO:
		/* get device's information */
		ixisa_copy_devinfo((ixisa_devinfo_t *) ioctl_param, dp);
		break;
	case IXISA_SET_SIG:
		/* set signal for ixisa interrupt */
		if (set_signal((ixisa_signal_t *) ioctl_param, dp))
			return FAILURE;
		break;
	case IXISA_READ_REG:
		/* read from register */
		if (read_reg((ixisa_reg_t *) ioctl_param, dp->base))
			return FAILURE;
		break;
	case IXISA_WRITE_REG:
		/* write to register */
		if (write_reg((ixisa_reg_t *) ioctl_param, dp->base))
			return FAILURE;
		if (((ixisa_reg_t *) ioctl_param)->id == IXISA_ADMCR)
			dp->admcr = ((ixisa_reg_t *) ioctl_param)->value;
		break;
	case IXISA_COUNT_CONF:
		/* set up 8254 counter */
		config_8254_count(dp->base,
						  ((ixisa_count_t *) ioctl_param)->id,
						  ((ixisa_count_t *) ioctl_param)->
						  id << 6 | _8254CW_RW_LMB | ((ixisa_count_t *)
													  ioctl_param)->mode,
						  ((ixisa_count_t *) ioctl_param)->value);
		break;
	case IXISA_AIDMA_CONF:
		/* configure and enable the dma transfer */
		dp->buf16 = (ixisa_buf16_t *) ioctl_param;
		dp->dma.mode = DMA_MODE_READ;

		disable_board_dma(dp);
		stop_board_pacer_count(dp->base);
		clear_board_irq(dp->base);

		if (config_dma(dp))
			return FAILURE;

		on_flag(&dp->flags, IXISA_FLAG_DMA);
		config_board_pacer_count(dp->base, dp->buf16->interval);
		enable_board_dma(dp->base);
		break;
	case IXISA_AIDMA_CONF_AUTOINIT:
		/* configure and enable the dma transfer with Autoinitialization mode */
		dp->buf16 = (ixisa_buf16_t *) ioctl_param;
		dp->dma.mode = DMA_MODE_READ | DMA_AUTOINIT;

		disable_board_dma(dp);
		stop_board_pacer_count(dp->base);
		clear_board_irq(dp->base);

		if (config_dma(dp))
			return FAILURE;

		on_flag(&dp->flags, IXISA_FLAG_DMA);
		config_board_pacer_count(dp->base, dp->buf16->interval);
		enable_board_dma(dp->base);
		break;
	case IXISA_AIDMA_NEXT:
		/* do next batch of the dma transfer directly */
		config_board_pacer_count(dp->base, dp->buf16->interval);
		enable_board_dma(dp->base);
		break;
	case IXISA_AIDMA_STOP:
		/* disable and stop the dma transfer */
		dp->buf16 = 0;
		disable_board_dma(dp);
		stop_board_pacer_count(dp->base);
		unconfig_dma(dp);
		off_flag(&dp->flags, IXISA_FLAG_DMA);
		break;
	case IXISA_AIDMA_HARVEST:
		if (get_flag(&dp->flags, IXISA_FLAG_DMA))
			memcpy(dp->buf16->buf, dp->dma.buf, dp->dma.sz);
		else
			return FAILURE;
		break;
	case IXISA_CLEAR_BOARD_IRQ:
		clear_board_irq(dp->base);
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int ixa826_release(struct inode *inode, struct file *file)
#else
void ixa826_release(struct inode *inode, struct file *file)
#endif
{
	/* (export)
	 *
	 * This function is called by ixisa.o whenever a process attempts to
	 * closes the device file. It doesn't have a return value in kernel
	 * version 2.0.x because it can't fail (you must always be able to
	 * close a device).  In version 2.2.x it is allowed to fail.
	 *
	 * Arguments: read <linux/fs.h> for (*release) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return 0;
	}

	dp->open = 0;				/* flag close */

	reset_board(dp);

	/* uninstall ISR */
	if (dp->irq)
		free_irq(dp->irq, dp);

	/* unconfigure dma */
	if (get_flag(&dp->flags, IXISA_FLAG_DMA)) {
		dp->buf16 = 0;
		unconfig_dma(dp);
		off_flag(&dp->flags, IXISA_FLAG_DMA);
	}

	/* release io region */
	release_region(dp->base, IO_RANGE);

	MOD_DEC_USE_COUNT;

# if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
# endif
}

static struct file_operations fops = {
	open:ixa826_open,
	release:ixa826_release,
	ioctl:ixa826_ioctl,
};

int ixa826_open(struct inode *inode, struct file *file)
{
	/* (export)
	 * This function is called by ixisa.o whenever a process attempts to open
	 * the device file of IXISA-A826
	 *
	 * Arguments: read <linux/fs.h> for (*open) of struct file_operations
	 *
	 * Returned:  none */

	int minor;
	ixisa_devinfo_t *dp;
	ixisa_signal_t sig;

	minor = MINOR(inode->i_rdev);
	dp = align_to_minor(minor);

	if (!dp) {
		/* if not found, get out */
		return -EINVAL;
	}

	++(dp->open);
	if (dp->open > 1) {
		/* if already opened by someone, get out */
		--(dp->open);
		return -EBUSY;
	}

	/* clear signal setting */
	sig.sid = 0;
	sig.pid = 0;
	sig.task = 0;
	set_signal(&sig, dp);

	/* request io region */
	request_region(dp->base, IO_RANGE, MODULE_NAME);

	/* reset board */
	reset_board(dp);

	/* if irq is configured, install */
	if (dp->irq)
		request_irq(dp->irq, irq_handler, SA_INTERRUPT | SA_SHIRQ,
					dp->name, dp);

	MOD_INC_USE_COUNT;

	return SUCCESS;
}

void cleanup_module()
{
	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s ", MODULE_NAME);

	/* uninstall file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = 0;
		off_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}
	printk(" has been removed.\n");
}

int init_module()
{
	/* initialize this module
	 *
	 * Arguments: none
	 *
	 * Returned:
	 *   integer 0 means ok, otherwise failed (module can't be load) */

	ixisa_devinfo_t *dp;

	printk(KERN_INFO FAMILY ": %s... ", MODULE_NAME);

	if (!ixisa_dev) {
		printk("device list is empty!\n");
		return FAILURE;
	}

	/* align to first IXISA-A826 in ixisa list */
	for (dev = ixisa_dev; dev && dev->id != A826; dev = dev->next);

	if (!dev) {
		printk("no %s card registered!\n", A826_NAME);
		return FAILURE;
	}

	/* install file operations */
	for (dp = dev; dp; dp = dp->next_f) {
		dp->fops = &fops;
		on_flag(&dp->flags, IXISA_FLAG_ATTEND);
		printk(".");
	}

	printk(" ready.\n");

	return SUCCESS;
}
