/* Read ADC result by pacer trigger and interrupt.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.3.1  2 Dec 2002 by Reed Lai
     Improves signal operating.

   v 0.3.0 29 Jan 2002 by Reed Lai
     Refine some register names.

   v 0.2.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.
   
   v 0.1.0 24 Jul 2001 by Reed Lai
     Update the name of Control Word Register from CWR to CR
  
   v 0.0.0 8 Mar 2001 by Reed Lai
     Create */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <signal.h>

#include "ixisa.h"

int fd;

void sig_handler(int sig)
{
	ixisa_reg_t reg;

	/* analog input */
	reg.id = IXISA_AI;
	reg.mode = IXISA_RM_NORMAL;

	if (ioctl(fd, IXISA_READ_REG, &reg))
		puts("Failure of analog input.");
	printf("%x ", reg.value);

	/* clear interrupt */
	reg.id = IXISA_CIR;
	if (ioctl(fd, IXISA_WRITE_REG, &reg))
		puts("Failure of clearing interrupt.");
}

int main()
{
	char *dev_file;
	ixisa_reg_t reg;
	ixisa_signal_t sig;
	static struct sigaction act, act_old;

	dev_file = "/dev/ixisa1";

	/* open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of open device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* signal action */
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);
	if (sigaction(SIGALRM, &act, &act_old)) {
		close(fd);
		puts("Failure of signal action.");
		return FAILURE;
	}

	/* signal condiction */
	sig.sid = SIGALRM;
	sig.pid = getpid();
	sig.is = 1;				/* INT_CHAN_0 */
	sig.bedge = 1;			/* signal for both edges */
	if (ioctl(fd, IXISA_SET_SIG, &sig)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of signal condiction.");
		return FAILURE;
	}

	/* pacer trigger, interrput, irq 5 */
	reg.id = IXISA_ADMCR;
	reg.value = 0x56;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of configuring interrupt.");
		return FAILURE;
	}

	/* gain for +/- 5V */
	reg.id = IXISA_ADGCR;
	reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of gain configuration.");
		return FAILURE;
	}

	/* channel 0 */
	reg.id = IXISA_ADMXCR;
	//reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of channel.");
		return FAILURE;
	}
	usleep(23);  /* delay 23us for channel switching */

	/* clear interrupt */
	reg.id = IXISA_CIR;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of clear interrupt.");
		return FAILURE;
	}

	/*****************************************
	 * setup 8254 counter-1/2 for INT_CHAN_3 *
	 *                                       *
	 * 2MHz / (0xffff * 0x1e) = 1Hz          *
	 *****************************************/
	reg.id = IXISA_8254CR;
	reg.value = 0x76;
		/* counter-1, mode 3 */

	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 1.");
		return FAILURE;
	}
	reg.id = IXISA_8254C1;
	reg.value = 0xff;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 1 LSB.");
		return FAILURE;
	}
	//reg.value = 0xff;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 1 MSB.");
		return FAILURE;
	}
	reg.id = IXISA_8254CR;
	reg.value = 0xb6;
		/* counter-2, mode 3 */

	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 2.");
		return FAILURE;
	}
	reg.id = IXISA_8254C2;
	reg.value = 0x1e;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 2 LSB.");
		return FAILURE;
	}
	reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of 8254 counter 2 MSB.");
		return FAILURE;
	}

	/* wait for exit */
	puts("Press <enter> to stop.");
	while (getchar() != '\n');

	close(fd);
	sigaction(SIGALRM, &act_old, NULL);
	puts("\nEnd of program.");
	return SUCCESS;
}
