/* Example of interrupt handle for DIO-24.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* 
   Enable the CN0PC3 interrupt.

   JP2
     +----------+
     |D   o o   |
     |          |
     |P   o o   |
     |  +-----+ |
     |E | o o | |
     |  +-----+ |
     +----------+

*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <signal.h>

#include "ixisa.h"

unsigned sig_counter;

void sig_handler(int sig)
{
	++sig_counter;
	putchar('+');
}

int main()
{
	int fd;
	char *dev_file;
	ixisa_reg_t reg;
	ixisa_signal_t sig;

	static struct sigaction act, act_old;

	sig_counter = 0;
	dev_file = "/dev/ixisa1";

	/* open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of openning device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* signal action */
	act.sa_handler = sig_handler;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGALRM);
	if (sigaction(SIGALRM, &act, &act_old)) {
		close(fd);
		puts("Failure of signal action.");
		return FAILURE;
	}

	/* Configure portC as input and portA B as output*/
	reg.id = IXISA_CR;
	reg.value = 0x89;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of PA.");
		return FAILURE;
	}

	/* signal condiction */
	sig.sid = SIGALRM;
	sig.pid = getpid();
	if (ioctl(fd, IXISA_SET_SIG, &sig)) {
		close(fd);
		sigaction(SIGALRM, &act_old, NULL);
		puts("Failure of signal condition.");
		return FAILURE;
	}

	/* rocking time */
	reg.id = IXISA_PA;
	puts("press several space keys to IRQ, then enter to get out\n");
	while (getchar() != 10) {
		reg.value = 0xff;  /* rising */
		if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
			close(fd);
			sigaction(SIGALRM, &act_old, NULL);
			puts("Failure of rising.");
			return FAILURE;
		}
		putchar('-');
	usleep(1000);
		reg.value = 0;  /* falling */
		if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
			close(fd);
			sigaction(SIGALRM, &act_old, NULL);
			puts("Failure of falling.");
			return FAILURE;
		}
		putchar('_');
	}
	printf("\nGet interrupt signal for %d times.\n", sig_counter);

	close(fd);
	sigaction(SIGALRM, &act_old, NULL);
	puts("\nEnd of program.");
	return SUCCESS;
}
