/* Example of I/O port for DIO-D48.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.1.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.

   v 0.0.0 17 Oct 2001 by Reed Lai
     create, blah blah... */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "ixisa.h"

int main()
{
	int fd;
	char *dev_file;
	ixisa_reg_t reg;
	unsigned int old = -1;

	dev_file = "/dev/ixisa1";

	/* Open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of openning device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* Configure CN1 as output */
	reg.id = IXISA_CN1CR;
	reg.value = 0x80;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		puts("Failure of writing register CN1CR!\n");
		close(fd);
		return FAILURE;
	}

	/* Configure CN2 as input */
	reg.id = IXISA_CN2CR;
	reg.value = 0x9b;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		puts("Failure of writing register CN2CR!\n");
		close(fd);
		return FAILURE;
	}

	/* Write CN1 ports, read CN2 ports */
	while (getchar() == 10) {
		printf("write CN1P[ABC]: [");
		reg.id = IXISA_CN1PA;
		++reg.value;
		if (!ioctl(fd, IXISA_WRITE_REG, &reg))
			printf("%2x", reg.value);
		else printf("--");
		
		reg.id = IXISA_CN1PB;
		++reg.value;
		if (!ioctl(fd, IXISA_WRITE_REG, &reg))
			printf(" %2x", reg.value);
		else printf("--");

		reg.id = IXISA_CN1PC;
		++reg.value;
		if (!ioctl(fd, IXISA_WRITE_REG, &reg))
			printf(" %2x]\n", reg.value);
		else printf("--]\n");

		printf("read  CN2P[ABC]: [");
		reg.id = IXISA_CN2PA;
		if (!ioctl(fd, IXISA_READ_REG, &reg))
			printf("%2x", reg.value);
		else printf("--");

		reg.id = IXISA_CN2PB;
		if (!ioctl(fd, IXISA_READ_REG, &reg))
			printf(" %2x", reg.value);
		else printf("--");

		reg.id = IXISA_CN2PC;
		if (!ioctl(fd, IXISA_READ_REG, &reg))
			printf(" %2x]\n", reg.value);
		else printf("--]\n");
		puts("Esc-enter to quit.");
	}

	close(fd);
	return SUCCESS;
}
