/* Read ADC result by software trigger.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.1.0 11 Jan 2002 by Reed Lai
     Rename DAQ to IXISA.
   
   v 0.0.0 7 Mar 2001 by Reed Lai
     Create */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "ixisa.h"

int main()
{
	int fd;
	char *dev_file;
	ixisa_reg_t reg, rai;
        float av;

	dev_file = "/dev/ixisa1";

	/* open device file */
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Failure of open device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	/* gain for +/- 5V */
	reg.id = IXISA_ADGCR;
	reg.value = 0;
	if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
		printf("Failure of ioctl command IXISA_WRITE_REG IXISA_ADGCR.\n");
		close(fd);
		return FAILURE;
	}

	/* read 32 channels by software trigger */
	rai.id = IXISA_AI;
	rai.mode = IXISA_RM_TRIGGER;
	reg.id = IXISA_ADMXCR;

	puts("Enter to continue, ESC to exit.\n");
	while (getchar() != 27) {
		for (reg.value = 0; reg.value < 32; ++reg.value) {
			if (ioctl(fd, IXISA_WRITE_REG, &reg)) {
				printf ("Failure of ioctl command IXISA_WRITE_REG IXISA_ADMXCR.\n");
				close(fd);
				return FAILURE;
			}
			usleep(10);
				/* delay 10us for channel switching */

			if (ioctl(fd, IXISA_READ_REG, &rai)) {
				printf ("Failure of ioctl command IXISA_WRITE_REG IXISA_AI.\n");
				close(fd);
				return FAILURE;
			}
			rai.value &= 0xfff; 	
			av = (( 10 * ((float)rai.value) / 0xfff) -5);
			printf("%x   %f\n", rai.value, av);
		}
	}

	close(fd);
	return SUCCESS;
}
