#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <curses.h>

#ifdef target_mservo3
#include "msw-mservo3.h"
#define MAX_CARDNO 10
#endif

#ifdef target_enc3
#include "msw-enc3.h"
#define MAX_CARDNO 1
#endif

#ifdef target_mstep2
#include "msw-mstep2.h"
#define MAX_CARDNO 20
#endif

#define LINE_SIZE 128
#define NAME_SIZE 16

typedef struct
{
	int fd;
	int base;
	int iorange;

}fdArray;
fdArray fds[MAX_CARDNO];	//0-MAX_CARDNO

#ifdef target_mservo3
int MSERVO3_INIT(unsigned char cardNo, unsigned int address)
{
	FILE *fd_proc;
	char dev_file[NAME_SIZE + 1];
	char rbuf[LINE_SIZE + 1];
	int fd_dev;
	ixisa_devinfo_t dev;
	ixisa_reg_t reg;
	int cardNum = 0;

	if (cardNo > MAX_CARDNO)
		return FAILURE;

	fd_proc = fopen(IXISA_PROC_FILE, "r");
	if (!fd_proc) {
		puts("There is not ISA device found.");
		return FAILURE;
	}

	while (fgets(rbuf, LINE_SIZE, fd_proc)) {
		if (!strncmp(rbuf, "dev:", 4)) {
			// make up available device file name 
			sprintf(dev_file, "/dev/%s", strtok(rbuf + 4, " "));
			
			// open device
			fd_dev = open(dev_file, O_RDWR);
			if (fd_dev < 0) {
				printf("Cannot open device file \"%s.\"\n", dev_file);
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}

			// get device information 
			if (ioctl(fd_dev, IXISA_GET_INFO, &dev)) {
				puts("Failure of getting device information.\n");
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}
			if (dev.id == SERVO300) {
				if (dev.base == address){
					fds[cardNo].fd = fd_dev;
					fds[cardNo].base = dev.base;
					fds[cardNo].iorange = 0x10;
					card[cardNo].base = dev.base;
					card[cardNo].op = 0;
					card[cardNo].workNo = 2;
					card[cardNo].sync1 = 0;
					card[cardNo].sync2 = 0;
					
					++cardNum;
					card[cardNo].exist = MSERVO3_EXIST(cardNo);
					break;
				}
			}else{ 
				close(fd_dev);
			}
		}
	}

	fclose(fd_proc);
	if (cardNum ==0) return FAILURE;
	WorkFlag = NO;
	state = PHASE1;
	sys.command = sIDLE;
	return card[cardNo].exist;
}
#endif

#ifdef target_enc3
int ENC3_INIT(unsigned int address)
{
	FILE *fd_proc;
	char dev_file[NAME_SIZE + 1];
	char rbuf[LINE_SIZE + 1];
	int fd_dev;
	ixisa_devinfo_t dev;
	ixisa_reg_t reg;

	fd_proc = fopen(IXISA_PROC_FILE, "r");
	if (!fd_proc) {
		puts("There is not ISA device found.");
		return FAILURE;
	}

	while (fgets(rbuf, LINE_SIZE, fd_proc)) {
		if (!strncmp(rbuf, "dev:", 4)) {
			// make up available device file name 
			sprintf(dev_file, "/dev/%s", strtok(rbuf + 4, " "));
			
			// open device
			fd_dev = open(dev_file, O_RDWR);
			if (fd_dev < 0) {
				printf("Cannot open device file \"%s.\"\n", dev_file);
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}

			// get device information 
			if (ioctl(fd_dev, IXISA_GET_INFO, &dev)) {
				puts("Failure of getting device information.\n");
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}
			if (dev.id == ENCODER300) {
				if (dev.base == address){
					fds[0].fd = fd_dev;
					fds[0].base = dev.base;
					fds[0].iorange = 0x10;
					break;
				}
			}else{ 
				close(fd_dev);
			}
		}
	}

	fclose(fd_proc);
	
	encoder_base = address;
	return SUCCESS;
	
}
#endif

#ifdef target_mstep2
int MSTEP2_INIT(unsigned char cardNo, unsigned int address)
{
	FILE *fd_proc;
	char dev_file[NAME_SIZE + 1];
	char rbuf[LINE_SIZE + 1];
	int fd_dev, fd;
	ixisa_devinfo_t dev;
	ixisa_reg_t reg;
	int cardNum = 0;

	if (cardNo > MAX_CARDNO)
		return FAILURE;

	fd_proc = fopen(IXISA_PROC_FILE, "r");
	if (!fd_proc) {
		puts("There is not ISA device found.");
		return FAILURE;
	}

	while (fgets(rbuf, LINE_SIZE, fd_proc)) {
		if (!strncmp(rbuf, "dev:", 4)) {
			//make up available device file name
			sprintf(dev_file, "/dev/%s", strtok(rbuf + 4, " "));

			// open device
			fd_dev = open(dev_file, O_RDWR);
			if (fd_dev < 0) {
				printf("Cannot open device file \"%s.\"\n", dev_file);
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}


			// get device information
			if (ioctl(fd_dev, IXISA_GET_INFO, &dev)) {
				puts("Failure of getting device information.\n");
				close(fd_dev);
				fclose(fd_proc);
				return FAILURE;
			}

			if (dev.id == STEP200) {
				if (dev.base == address){
					fd = fd_dev;
					fds[cardNo].fd = fd_dev;
					fds[cardNo].base = dev.base;
					fds[cardNo].iorange = 0x10;
					card[cardNo].base = dev.base;
					card[cardNo].DDA_cycle = d_DDA_cycle;
					card[cardNo].Acc_Dec = d_Acc_Dec;
					card[cardNo].Low_Speed = d_Low_Speed;
					card[cardNo].High_Speed = d_High_Speed;

				}
			}else{
				close(fd_dev);
			}
		}
	}

	fclose(fd_proc);
	High_Speed = d_High_Speed;
	output_fifo(cardNo, TEST_FIFO, 0x84, 0, 0, 0, 0, 0, 0, 0);
	usleep(500000);	//waiting for response
	
	reg.id = IXISA_STS;
	ioctl(fd, IXISA_READ_REG, &reg);
	if (reg.value !=84)
		card[cardNo].exist = NO;
	else
		card[cardNo].exist = YES;

	WorkFlag = NO;
	state = PHASE1;
	sys.command = sIDLE;
	return card[cardNo].exist;
}
#endif
					
unsigned int myinb(unsigned int cdata)
{
	ixisa_reg_t reg;
	int x;
	x=0;
	for (x = 0; x < MAX_CARDNO; x++)
	{
		if ((cdata >= fds[x].base) && (cdata < (fds[x].base + fds[x].iorange)))
		{
			switch (cdata - fds[x].base) {
#ifdef target_mservo3
				case FIFO:
					reg.id = IXISA_FIFO;
					if (ioctl(fds[x].fd, IXISA_READ_REG, & reg))
						puts("Failure of IXISA_READ_REG:FIFO\n");
					break;

				case MR:
					reg.id = IXISA_MR;
					if (ioctl(fds[x].fd, IXISA_READ_REG, & reg))
						puts("Failure of IXISA_READ_REG:MR\n");
					break;
		
				case DI:
					reg.id = IXISA_DI;
					if (ioctl(fds[x].fd, IXISA_READ_REG, & reg))
						puts("Failure of IXISA_READ_REG:DI\n");
					break;
#endif
#if target_enc3
				case VALUE1:
					reg.id = IXISA_VALUE1;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:VALUE1\n");
					break;

				case VALUE2:
					reg.id = IXISA_VALUE2;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:VALUE2\n");
					break;

				case VALUE3:
					reg.id = IXISA_VALUE3;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:VALUE3\n");
					break;

				case INDEX:
					reg.id = IXISA_INDEX;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:INDEX\n");
					break;
#endif
#ifdef target_mstep2
				case DI1:
					reg.id = IXISA_DI1;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:DI1\n");
					break;

				case DI2:
					reg.id = IXISA_DI2;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:DI2\n");
					break;
					
				case STS:
					reg.id = IXISA_STS;
					if (ioctl(fds[x].fd, IXISA_READ_REG, &reg))
						puts("Failure of IXISA_READ_REG:STS\n");
					break;
#endif
					
				default:
					return FAILURE;
			}

			break;
		}
		++x;

	}
	return reg.value;

}

void myoutb(unsigned int base, unsigned char cdata)
{
	ixisa_reg_t reg;
	int x, match;

	reg.value = cdata;

	for (x = 0; x < MAX_CARDNO; x++){
		if((base >= fds[x].base) && (base < (fds[x].base + fds[x].iorange))){
			switch (base - fds[x].base){
#if 0
				case TJ_CNTL:
					reg.id = IXISA_TJ_CNTL;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg)){
						puts("Failure of IXISA_WRITE_REG:TJ_CNTL");
					}
					break;

				case TJ_AUXC:
					reg.id = IXISA_TJ_AUXC;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg)){
						puts("Failure of IXISA_WRITE_REG:TJ_AUXC");
					}
					break;
#endif			
#ifdef target_mservo3
				case DO:
					reg.id = IXISA_DO;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG\n");
					break;
					
				case FIFO:
					reg.id = IXISA_FIFO;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg)){
						puts("Failure of IXISA_WRITE_REG:FIFO\n");
					}
					break;

				case RR:
					reg.id = IXISA_RR;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg)){
						puts("Failure of IXISA_WRITE_REG:RR\n");
					}
					break;
				
#endif
#ifdef target_enc3
				case CTRL1:
					reg.id = IXISA_CTRL1;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG:CTRL1\n");
					break;

				case CTRL2:
					reg.id = IXISA_CTRL2;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG:CTRL2\n");
					break;

				case CTRL3:
					reg.id = IXISA_CTRL3;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG:CTRL3\n");
					break;
#endif	
#ifdef target_mstep2
				case FIFO:
					reg.id = IXISA_FIFO;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG:FIFO\n");
					break;

				case RR:
					reg.id = IXISA_RR;
					if (ioctl(fds[x].fd, IXISA_WRITE_REG, &reg))
						puts("Failure of IXISA_WRITE_REG:RR\n");
					break;
#endif

				default:
					puts("FAILURE\n");
					break;
			}

			break;
		}
	}
}

void mymove(int x, int y)
{
	/* FIXME - gotoxy, the function of c++, not supported yet */
	printf("\nLeft:%x, Top:%x\n", x, y);
/*	initscr();
	refresh();
	move(y, x);
	printw("test\n");
	refresh();
*/
}

void closeDev(){
	unsigned char j;
	for ( j = 0; j < MAX_CARDNO; j++) {
		close(fds[j].fd);
	}
}
