/* Add card to the list.

   Author: Reed Lai

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/* File level history (record changes for this file here.)

   v 0.2.1  4 Apr 2003 by Reed Lai
     Message changement.

   v 0.2.0  1 Mar 2002 by Reed Lai
     Add a parameter for DMA.

   v 0.1.0 11 Jan 2001 by Reed Lai
     Rename DAQ to IXISA.

   v 0.0.0 by Reed Lai 26 Feb 2001
     Create */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "ixisa.h"

enum {
	COMMAND,
	CARD_NO,
	CARD_NAME,
	CARD_DMA,
	CARD_IRQ,
	CARD_BASE,
	ARGC
};

int main(int argc, char *argv[])
{
	int fd;
	char *dev_file;
	ixisa_devinfo_t dev;

	if (argc != ARGC)
		return FAILURE;

	strncpy(dev.name, argv[CARD_NAME], CNL);
	dev.no = (unsigned int) atoi(argv[CARD_NO]);
	dev.dma.nr = (unsigned int) atoi(argv[CARD_DMA]);
	dev.irq = (unsigned int) atoi(argv[CARD_IRQ]);
	dev.base = (unsigned long) strtol(argv[CARD_BASE], 0, 16);

	printf("Add card #%u for %s dma=%u irq=%u base=0x%lx ... ", dev.no,
		   dev.name, dev.dma.nr, dev.irq, dev.base);

	dev_file = "/dev/ixisa";
	fd = open(dev_file, O_RDWR);
	if (fd < 0) {
		printf("Can not  open device file \"%s.\"\n", dev_file);
		return FAILURE;
	}

	if (ioctl(fd, IXISA_ADD_DEV, &dev)) {
		printf("Failure of ioctl command IXISA_ADD_DEV.\n");
		close(fd);
		return FAILURE;
	}

	close(fd);
	printf("done.\n");

	return SUCCESS;
}
